*---------- Location Section ----------
*} Program: AddInfo.prg
*--------------------------------------
*) Description: 6.0 Class Browser AddIn that fills the right-side tree with
*)  additional information about methods and properties.
*---------- Usage Section -------------
*$ Parameters: toBrowser - instance of class browser calling this addin
*$ Usage: oBrowser.AddIn(cName, cProgram, cMethod)
*$ Example: _oBrowser.AddIn('AddInfo', 'c:\vfptools\AddInfo.prg', 'RefreshMembers')
*$ Returns: LOGICAL 
*--------------------------------------
#INCLUDE addinfo.h

********************************
****** Parameters ********
lparameters toBrowser
********************************

local llOldLockScreen
local lnOldMousePointer
local lnOldWorkArea
local lnRecordNumber
local lcViewCodeFileName
local lcMethodsField
local lcProtectedField
local lcPropertiesField
local loNode
local lnIndex
local lnCount
local lcItem
local lcProcedureName
local lnProcedureCount
local llIsProtected
local lcEqualTo
local lcNodeText
local lnPropertyCount
local lnEndOfPropName
local lcPropertyName 
local lcNodeVariable
local lcNodeVariableTemp

* Do not run if any of these are true:
* 1) Parameter is not an object
* 2) File is chosen in the left side, not a class
* 3) Right-side tree "oleMembers" is not an object.
*      This probably means we are in "List" mode
* 4) No files are opened
if vartype(toBrowser) # cc_OBJECT_TYPE ;
	or toBrowser.lFileMode = cc_CLASS_NOT_SELECTED ;
	or type('toBrowser.oleMembers') # cc_OBJECT_TYPE ;
	or empty(toBrowser.cAlias)
	return
endif

* Check to see if the menu AddIn has disabled us
lnOldWorkArea = select(cc_CURRENT_WORKAREA)
SELECT browser

lnRecordNumber = recno()
locate for upper(alltrim(TYPE)) == 'ADDIN' ;
			and upper(alltrim(ID)) == 'METHOD' ;
			and upper(alltrim(NAME)) == program() ;
			and upper(alltrim(METHOD)) == 'REFRESHMEMBERS'
if not found() ;
	or upper(alltrim(USER1)) == cc_BROWSER_USER1_DISABLED
	* since we are disabled put things back and get out
	
	if between(lnRecordNumber, 1, reccount())
		goto lnRecordNumber
	endif

	select (lnOldWorkArea)
	return
endif

* move the record pointer back in the browser cursor since
* we probably moved it with the locate above
if between(lnRecordNumber, 1, reccount())
	goto lnRecordNumber
endif

* Get in the correct Metadata workarea for the
*  selected class
SELECT (toBrowser.cAlias)

* Look for the MetaData record that holds
* the methods and properties for the
* selected class
lnRecordNumber = recno()
locate for alltrim(lower(ObjName)) == alltrim(lower(toBrowser.cClass)) ;
	and ! empty(TIMESTAMP)
	
if not found()
	* We didn't find it so there's no work for us to do.
	*  (BTW, this shouldn't happen <s>)
	* Put things back and get out.
	if between(lnRecordNumber, 1, reccount())
		goto lnRecordNumber
	endif
	select (lnOldWorkArea)
	return
endif && not found()

* Save the current state of the environment and
*  set the display to look like we're doing something
llOldLockScreen = toBrowser.lockscreen
toBrowser.lockscreen = .t.
* Making the treeviews invisible makes the mouse cursor
* an hourglass even over the trees
toBrowser.oleMembers.Visible = .f.
toBrowser.oleClassList.Visible = .f.

lnOldMousePointer = _screen.mousepointer
_screen.mousepointer = cc_MOUSE_HOURGLASS

* Grab the data we need from the Memo fields
lcMethodsField = Methods
lcPropertiesField = Properties
lcProtectedField = Protected

* Done with the cursor, so put things back
if between(lnRecordNumber, 1, reccount())
	goto lnRecordNumber
endif
select (lnOldWorkArea)

*** METHODS ***
* Find the parent node for all methods
*  BTW, we don't return the node object here
*  because that causes GPF problems (you know...00000005)
*  But, get the index first and using it to access the node
*  is ok.  Probably a problem with the stack management
*  internal to VFP when it has ActiveX objects on the
*  stack.
lnIndex = GetMainNode(toBrowser, cc_METHODS_NODE_NAME)
if lnIndex>0
	lcNodeVariable = sys(2015)
	local &lcNodeVariable.
	&lcNodeVariable. = toBrowser.oleMembers.object.Nodes(lnIndex).Child
	*** Key and Text properties are the same for all nodes...
	do while type(lcNodeVariable+'.Text') = cc_CHARACTER_TYPE
		if not cc_EMPTY_MARK $ &lcNodeVariable..Text
			&lcNodeVariable..Bold = .t.
		endif
		lcNodeVariableTemp = lcNodeVariable
		lcNodeVariable = sys(2015)
		local &lcNodeVariable.
		&lcNodeVariable. = &lcNodeVariableTemp..Next
	enddo
endif

* Look for procedures (methods) defined for
*  this class.  The methods field of the
*  .VCX table contains the source code.
_mline=0
lnProcedureCount=1
FOR lnCount = 1 TO MEMLINE(lcMethodsField)
	lcItem=UPPER(ALLTRIM(MLINE(lcMethodsField, 1, _mline)))
	IF EMPTY(lcItem)
		LOOP
	ENDIF

	if left(lcItem,LEN(cc_PROC_KEYWORD)) == cc_PROC_KEYWORD
		lcProcedureName = substr(lcItem, LEN(cc_PROC_KEYWORD)+1)
	else
		loop
	endif

	* Put each one we find into an array.
	*  We'll sort these later...
	dimension aMethods[lnProcedureCount]
	aMethods[lnProcedureCount] = lcProcedureName
	lnProcedureCount = lnProcedureCount + 1
endfor

* If we found any methods for the class, sort them...
if type('aMethods') # cc_UNKNOWN_TYPE
	asort(aMethods)
	
	* Then add to the tree methods not already there
	FOR lnProcedureCount = 1 TO alen(aMethods)
		lnIndex = IsItemInMembersTree(aMethods[lnProcedureCount], ;
			toBrowser, cc_METHODS_NODE_NAME)
		if not (lnIndex>0)
			* Protected methods get a different icon
			llIsProtected = (ATC(CR ;
					+ ltrim(upper(aMethods[lnProcedureCount])) + CR, ;
					CR + upper(lcProtectedField)) # 0) ;
				OR 	(ATC(LF + ltrim(upper(aMethods[lnProcedureCount])) + ;
					LF, LF + upper(lcProtectedField)) # 0)
			if not (llIsProtected and !toBrowser.lProtectedFilter)
				AddItemToMembers(aMethods[lnProcedureCount], ;
					toBrowser, cc_METHODS_NODE_NAME,  llIsProtected)
			endif
		endif
	ENDFOR
endif	


*** PROPERTIES ***

* Find the parent node for all properties.  Same trick
*  with the node index to avoid the GPF.
lnIndex = GetMainNode(toBrowser, cc_PROPERTIES_NODE_NAME)
* Properties that have been changed from their default values
*  will appear in the .VCX's Properties field.  If we find them
*  in there, make them bold in the tree.
if lnIndex>0
	lcNodeVariable = sys(2015)
	local &lcNodeVariable.
	&lcNodeVariable. = toBrowser.oleMembers.object.Nodes(lnIndex).Child
	*** Key and Text properties are the same for all nodes...
	do while type(lcNodeVariable+'.Text') = cc_CHARACTER_TYPE
		lcNodeText = ltrim(upper(chrtranc(&lcNodeVariable..Text,cc_ALL_MARKS,"")))
		if ATC(CR+lcNodeText, CR+upper(lcPropertiesField)) # 0 OR ;
			ATC(LF+lcNodeText, LF+upper(lcPropertiesField)) # 0
			&lcNodeVariable..Bold = .t.
		endif
		lcNodeVariableTemp = lcNodeVariable
		lcNodeVariable = sys(2015)
		local &lcNodeVariable.
		&lcNodeVariable. = &lcNodeVariableTemp..Next
	enddo
endif

* Look through the properties field and make an
*  array of all the names we find.  We'll sort these
*  later...
_mline=0
lnPropertyCount=1
FOR lnCount = 1 TO MEMLINE(lcPropertiesField)
	lcItem=lower(ALLTRIM(MLINE(lcPropertiesField, 1, _mline)))
	IF EMPTY(lcItem)
		LOOP
	ENDIF

	lnEndOfPropName = ATC('=', lcItem)
	if lnEndOfPropName # 0
		lcPropertyName = alltrim(left(lcItem, lnEndOfPropName - 1))
	else
		lcPropertyName = lcItem
	endif && lnEndOfPropName # 0

	dimension aProperties[lnPropertyCount, 2]
	aProperties[lnPropertyCount, cc_PROPERTY_NAME] = lcPropertyName 
	aProperties[lnPropertyCount,cc_PROPERTY_VALUE] = lcItem
	lnPropertyCount=lnPropertyCount + 1
endfor

* Add the value of the property to the end of
*  it's name in the tree.  This makes it easy to
*  glance and see them, just like the properties
*  sheet.
if type('aProperties') # cc_UNKNOWN_TYPE
	* First sort the properties so they are easy to find
	*  in the tree.
	asort(aProperties)
	FOR lnPropertyCount = 1 TO alen(aProperties, 1)
		lnIndex = 	IsItemInMembersTree( ;
			aProperties[lnPropertyCount, cc_PROPERTY_NAME], ;
			toBrowser, cc_PROPERTIES_NODE_NAME)
		if lnIndex <= 0
			* Add to the tree non-default properties from superclasses
			* and base class.
			
			* If the property is protected, we'll use a different icon
			*in the tree
			llIsProtected = (ATC(CR + ltrim(upper( ;
					aProperties[lnPropertyCount, cc_PROPERTY_NAME])), ;
					CR + upper(lcProtectedField)) # 0) ;
				OR ;
					(ATC(LF + ltrim(upper( ;
					aProperties[lnPropertyCount, cc_PROPERTY_NAME])), ;
					LF+upper(lcProtectedField)) # 0)
			lnIndex = AddItemToMembers( ;
				aProperties[lnPropertyCount, cc_PROPERTY_NAME], ;
				toBrowser, cc_PROPERTIES_NODE_NAME, llIsProtected)
			lcEqualTo = aProperties[lnPropertyCount, cc_PROPERTY_VALUE]
			
			* Properties that are there with nothing to the right
			*  of the "=" look like "(None)" on the property sheet.
			if right(rtrim(lcEqualTo),1) = '='
				lcEqualTo = lcEqualTo + cc_EMPTY_PROPERTY_VALUE
			endif
			toBrowser.oleMembers.Object.Nodes(lnIndex).Text = lcEqualTo
		else
			* For the properties that are already in the tree, 
			*just add thier values to their right.
			lcEqualTo = right( ;
				aProperties[lnPropertyCount, cc_PROPERTY_VALUE], ;
				len(aProperties[lnPropertyCount, cc_PROPERTY_VALUE]) ;
				- len(aProperties[lnPropertyCount, cc_PROPERTY_NAME]))
			if right(rtrim(lcEqualTo),1) = '='
				lcEqualTo = lcEqualTo+cc_EMPTY_PROPERTY_VALUE
			endif
			with toBrowser.oleMembers.Object.Nodes(lnIndex)
				.Text = .Text + lcEqualTo
			endwith
		endif
	ENDFOR
endif

* Put stuff back the way we found it.
_screen.mousepointer = lnOldMousePointer
toBrowser.oleMembers.Visible = .t.
toBrowser.oleClassList.Visible = .t.
toBrowser.lockscreen = llOldLockScreen

* QED
return

***-------------------------------------------***
function AddItemToMembers(tcMethod, toBrowser, tcParent, tlProtected)
***-------------------------------------------***
	local lcItemIcon, lcItemType
	do case 
		case tcParent == cc_METHODS_NODE_NAME
			if tlProtected
				lcItemIcon = "methodc.ico"
			else
				lcItemIcon = "method.ico"
			endif
			lcItemType = "m"
		case tcParent == cc_PROPERTIES_NODE_NAME
			if tlProtected
				lcItemIcon = "propc.ico"
			else
				lcItemIcon = "prop.ico"
			endif
			lcItemType = "p"
	endcase

	local loMethodNode
	if GetMainNode(toBrowser, tcParent) <= 0
		local loNode
		loNode = toBrowser.oleMembers.Object.Nodes.Add(, cc_TVWCHILD, ;
			tcParent, tcParent)
		IF toBrowser.lAutoExpand
			loNode.Expanded = .T.
		ENDIF
		WITH toBrowser
			.nMemberCount = .nMemberCount + 1
			DIMENSION .aMemberList[.nMemberCount, 3]
			.aMemberList[.nMemberCount, 1] = tcParent
			.aMemberList[.nMemberCount, 2] = ""
			.aMemberList[.nMemberCount, 3] = ""
		ENDWITH
	ENDIF

	lcItem = lower(tcMethod)
	with toBrowser
		loNode = .oleMembers.Object.Nodes.Add(tcParent, ;
			cc_TVWCHILD, lcItem, lcItem, , )
		loNode.Bold = .t.
		loNode.ForeColor = RGB(cc_ADDED_NODE_COLOR)
		local lnImageIndex
		lnImageIndex = .GetImageIndex(.T., lcItemIcon)

		IF lnImageIndex > 0
			loNode.Image = ALLTRIM(.aImageList[lnImageIndex])
		ENDIF
	endwith
	*** what does Ken know that we don't about spaces in method names???
	local lnAtPos
	lnAtPos = AT(" ", lcItem)
	IF lnAtPos > 0
		lcItem = ALLTRIM(SUBSTR(lcItem, lnAtPos + 1))
	ENDIF
	WITH toBrowser
		.nMemberCount = .nMemberCount + 1
		DIMENSION .aMemberList[.nMemberCount,3]
		.aMemberList[.nMemberCount,1] = ALLTRIM(CHRTRAN(lcItem, ;
			cc_ALL_MARKS, ""))
		.aMemberList[.nMemberCount,2] = lcItemType
		.aMemberList[.nMemberCount,3] = ""
	ENDWITH
	return loNode.index
endfunc 

***-------------------------------------------***
function IsItemInMembersTree(tcMethod,toBrowser,tcParent)
***-------------------------------------------***
	local loMethodNode
	local lnIndex
	local loChildNode
	lnIndex = GetMainNode(toBrowser, tcParent)
	if vartype(lnIndex) # 'N' or lnIndex<=0
		return 0
	endif

	local lcNodeVariable
	local lcNodeVariableTemp

	lcNodeVariable = sys(2015)
	local &lcNodeVariable.
	&lcNodeVariable. = toBrowser.oleMembers.object.Nodes(lnIndex).Child
	*** Key and Text properties are the same for all nodes...
	do while type(lcNodeVariable+'.Text') = cc_CHARACTER_TYPE
		if upper(tcMethod) = upper(ltrim(chrtranc(&lcNodeVariable..Text,cc_ALL_MARKS,'')))
			return &lcNodeVariable..Index
		endif
		lcNodeVariableTemp = lcNodeVariable
		lcNodeVariable = sys(2015)
		local &lcNodeVariable.
		&lcNodeVariable. = &lcNodeVariableTemp..Next
	enddo
	
	return 0
endfunc &&* IsInMembersTree

***-------------------------------------------***
function GetMainNode(toBrowser, tcName)
***-------------------------------------------***
	local lcNodeVariable
	local lcNodeVariableTemp

	lcNodeVariable = sys(2015)
	local &lcNodeVariable.
	&lcNodeVariable. = toBrowser.oleMembers.Object.Nodes(1)
	*** Key and Text properties are the same for all nodes...
	do while type(lcNodeVariable+'.Key') = cc_CHARACTER_TYPE
		if &lcNodeVariable..Key = tcName
			return &lcNodeVariable..Index
		endif
		lcNodeVariableTemp = lcNodeVariable
		lcNodeVariable = sys(2015)
		local &lcNodeVariable.
		&lcNodeVariable. = &lcNodeVariableTemp..Next
*		release &lcNodeVariableTemp.
	enddo

	return -1
endfunc &&* GetMainNode

*---------- Maintenance Section -------
*@ Precondition Invariants: 
*@ Post-condition Invariants: 
*? To Do:
*= Methods Called:
** Process: 
*^ Change Log: 
*^ Created 20:03:57 Friday, March 19, 1999 - NDD
*---------- (c) Copyright 1999 ----
*   Company: East West Technology Services, Inc.
*   Programmer: Nick Drochak
*   Internet: ndrochak@ewc.co.jp
*   Phone: 81-3-3222-5912
*--------------------------------------